	; PC-FX example program
	; by Matej Horvat

	; Demonstrates KING backgrounds, printing text, reading the pad, and
	; using 7up sprites.

	!ORG 0x8000

	; ----------------------------------------

Start:
	; Point the stack pointer to the end of RAM
	?mov 0x200000, $SP

	; Tetsu initialization

	; Tetsu screen mode
	mov 0, $6	; Screen mode
	?mov 0xE0D, $7	; 320*262/int.sync/16&16 colors/7up spr + KING BG0&1
	jal pcfxTetsuWrite

	; Tetsu palette
	mov 1, $6	; Palette number
	mov 0, $7
	jal pcfxTetsuWrite
	?mov 16, $20
	?mov .tetsuPalette, $21
	mov 2, $6	; Palette data W
.nextTetsuPaletteEntry:
	ld.h 0[$21], $7
	jal pcfxTetsuWrite
	add 2, $21
	add -1, $20
	bnz .nextTetsuPaletteEntry

	; 7up palette offsets
	mov 4, $6	; 7up palette offsets
	mov 0, $7	; BG & spr. offset = 0
	jal pcfxTetsuWrite

	; KING BG0/1 palette offsets
	mov 5, $6	; KING BG0/1 palette offsets
	mov 0, $7	; Both offsets = 0
	jal pcfxTetsuWrite

	; Priority: 7up spr = 3, KING BG0 = 2, KING BG1 = 1
	; Lower priorities go in the back.
	mov 8, $6	; 7up/Rainbow priority
	?mov 0x30, $7	; 7up spr = 3
	jal pcfxTetsuWrite
	mov 9, $6	; KING priority
	?mov 0x12, $7	; BG0 = 2, BG1 = 1
	jal pcfxTetsuWrite

	; Cellophane control: disable
	?mov 0xE, $6	; Cellophane control
	mov 0, $7	; Disable
	jal pcfxTetsuWrite

	; KING initialization

	; Initialize BG0 KRAM
	; (0 to 0x10000 - see BG0 CG address)
	?mov 0x10000, $20
	mov 0, $21
.nextKRAMWrite1:
	mov 0xD, $6	; KRAM write address
	mov $21, $7	; Address
	jal pcfxKingWriteW
	mov 0xE, $6	; KRAM R/W
	mov 4, $7	; Value
	jal pcfxKingWriteH
	add 1, $21
	add -1, $20
	bnz .nextKRAMWrite1

	; Print some text
	mov 4, $6
	mov 4, $7
	?mov .text, $8
	jal PrintStr

	; Initialize BG1 KRAM
	; (0x10000 to 0x14000 - see BG1 CG address)
	?mov 0x4000, $20
	?mov 0x10000, $21
.nextKRAMWrite2:
	mov 0xD, $6	; KRAM write address
	mov $21, $7	; Address
	jal pcfxKingWriteW
	mov 0xE, $6	; KRAM R/W
	?mov 0x1111, $7	; Value
	jal pcfxKingWriteH
	add 1, $21
	add -1, $20
	bnz .nextKRAMWrite2

	; KRAM page setup: use page 0 for everything
	; (we only use backgrounds anyway)
	mov 0xF, $6	; KRAM page setup
	mov 0, $7	; Everything = page 0
	jal pcfxKingWriteW

	; BG mode: BG0 = 4-color palette, BG1 = 4-color palette
	?mov 0x10, $6	; BG mode
	?mov 0x11, $7	; BG0 = 4-color palette, BG1 = 4-color palette
	jal pcfxKingWriteH

	; BG priority: BG1 = backmost, BG0 in front of it, others hidden
	?mov 0x12, $6	; BG priority
	?mov 0xA, $7	; (binary 001 010)
	jal pcfxKingWriteH

	; KING microprogram
	?mov 0x15, $6	; Microprogram control
	mov 0, $7	; Running = off
	jal pcfxKingWriteH
	?mov 0x13, $6	; Microprogram write address
	mov 0, $7	; Address
	jal pcfxKingWriteH
	?mov .kingMicroprogram, $20
	?mov 16, $21
.microprogramLoop:
	?mov 0x14, $6	; Microprogram data
	ld.h 0[$20], $7
	jal pcfxKingWriteH
	add 2, $20
	add -1, $21
	bnz .microprogramLoop
	?mov 0x15, $6	; Microprogram control
	mov 1, $7	; Running = on
	jal pcfxKingWriteH

	; BG scroll mode: BG0/1 = single background area
	?mov 0x16, $6	; BG scroll mode
	mov 0, $7	; BG0/1 = single background area
	jal pcfxKingWriteH

	; BG0 CG address: it starts at 0
	?mov 0x21, $6	; BG0 CG address
	mov 0, $7
	jal pcfxKingWriteH

	; BG1 CG address: it starts at 0x10000
	?mov 0x25, $6	; BG1 CG address
	?mov 64, $7
	jal pcfxKingWriteH

	; BG0 size: 512*512
	?mov 0x2C, $6	; BG0 size
	?mov 0x99, $7	; 512*512
	jal pcfxKingWriteH

	; BG1 size: 256*256
	?mov 0x2D, $6	; BG1 size
	?mov 0x88, $7	; 256*256
	jal pcfxKingWriteH

	; BG0 X/Y scroll: 0, -16
	?mov 0x30, $6	; BG0 X scroll
	mov 0, $7
	jal pcfxKingWriteH
	?mov 0x31, $6	; BG0 Y scroll
	mov -16, $7
	jal pcfxKingWriteH

	; BG1 X/Y scroll: 0, 0
	?mov 0x32, $6	; BG1 X scroll
	mov 0, $7
	jal pcfxKingWriteH
	?mov 0x33, $6	; BG1 Y scroll
	mov 0, $7
	jal pcfxKingWriteH

	; Pad 0 init

	out.h $0, 0[$0]
	out.w $0, 0x40[$0]
	in.h 0[$0], $0	; Required(?) dummy read

	; 7up A init
	; The order of setting the registers is important (but unknown),
	; copied here from Eris.

	; The values were obtained experimentally from Mednafen
	; and are meant for Tetsu resolution 320*262.
	; The top-left coordinates for a sprite with those settings are
	; 0x1D for X and 0x40 for Y.
	mov 0xA, $6
	?mov 0x0400, $7	; Eris and MagicKit say 0x202
	jal pcfx7upAWrite
	mov 0xB, $6
	?mov 0x0629, $7	; Eris says 0x41F, MagicKit says 0x31F
	jal pcfx7upAWrite
	mov 0xC, $6
	?mov 0x1104, $7	; Eris says 0x1102, MagicKit says 0xF02
	jal pcfx7upAWrite
	mov 0xD, $6
	?mov 0xEF, $7	; Eris and MagicKit say 0xEF
	jal pcfx7upAWrite
	mov 0xE, $6
	mov 0, $7	; Eris says 2, MagicKit says 3
	jal pcfx7upAWrite

	; Enable SAT DMA
	mov 0xF, $6	; DMA control
	?mov 0x10, $7	; "VRAM-SATB transfer repetition"
	jal pcfx7upAWrite

	; SAT address: 0xFF00
	?mov 0x13, $6	; SAT address
	?mov 0xFF00, $7
	jal pcfx7upAWrite

	; Control: display sprite
	mov 5, $6	; Control
	?mov 0x40, $7	; Display sprite
	jal pcfx7upAWrite

	; Sprite pattern: fill the first 256 halfwords with some data
	mov 0, $6	; Memory write address
	mov 0, $7
	jal pcfx7upAWrite
	?mov 256, $20
	mov 2, $6	; Memory write data
	?mov 0x1111, $7
.spritePatternLoop:
	jal pcfx7upAWrite
	add 4, $7
	add -1, $20
	bnz .spritePatternLoop

	?mov 0xB0, $20
	?mov 0xB0, $21
.loop:
	jal pcfxTetsuWaitFrame

	; Sprite 0
	mov 0, $6	; Memory write address
	?mov 0xFF00, $7	; Starting address
	jal pcfx7upAWrite
	mov 2, $6	; Memory write data
	mov $20, $7	; Y
	jal pcfx7upAWrite
	mov $21, $7	; X
	jal pcfx7upAWrite
	mov 0, $7	; Sprite data address / 32
	jal pcfx7upAWrite
	?mov 0x1180, $7	; Height = 32, width = 32, priority = 1, palette = 0
	jal pcfx7upAWrite

	jal pcfxReadPad0
	andi 0x100, $10, $0	; Up?
	bz .checkPadDown
	add -1, $20
.checkPadDown:
	andi 0x400, $10, $0	; Down?
	bz .checkPadLeft
	add 1, $20
.checkPadLeft:
	andi 0x800, $10, $0	; Left?
	bz .checkPadRight
	add -1, $21
.checkPadRight:
	andi 0x200, $10, $0	; Right?
	bz .loop
	add 1, $21
	br .loop

.tetsuPalette:
	; EGA/VGA colors 0-15
	?dh 0x88
	?dh 0xFC7
	?dh 0x4B54
	?dh 0x5A94
	?dh 0x266C
	?dh 0x35AB
	?dh 0x7148
	?dh 0xC088
	?dh 0x8088
	?dh 0x1DF6
	?dh 0x9621
	?dh 0xB3A0
	?dh 0x4C5F
	?dh 0x69DE
	?dh 0xE209
	?dh 0xFF88

.kingMicroprogram:
	?dh 0x0000	; "+0 from CG is accessed", BG0
	?dh 0x0040	; "+0 from CG is accessed", BG1
	?dh 0x0041	; "+1 from CG is accessed", BG1
	?dh 0x0100	; NOP
	?dh 0x0100	; NOP
	?dh 0x0100	; NOP
	?dh 0x0100	; NOP
	?dh 0x0100	; NOP
	?dh 0x0100	; NOP
	?dh 0x0100	; NOP
	?dh 0x0100	; NOP
	?dh 0x0100	; NOP
	?dh 0x0100	; NOP
	?dh 0x0100	; NOP
	?dh 0x0100	; NOP
	?dh 0x0100	; NOP

.text:
	?cstring "PC-FX example"

	; ----------------------------------------

pcfxTetsuWrite:
	; Writes a halfword to a Tetsu register.

	; IN:
	; $6 = register
	; $7 = value

	out.h $6, 0x300[$0]
	out.h $7, 0x304[$0]

	jmp $LP

	; ----------------------------------------

pcfxTetsuWaitFrame:
	; Waits until the start of the next video frame.

	?mov 0x20A0, $12	; Last raster
	?mov 0x3FE0, $13	; Raster mask

	; First, wait until Tetsu is finished with the current frame.
.readAgain1:
	in.h 0x300[$0], $10
	in.h 0x300[$0], $11	; Read twice because of hardware bug
	and $13, $10		; Get the raster
	and $13, $11
	cmp $10, $11
	bne .readAgain1
	cmp $10, $12
	bne .readAgain1

	; Then wait until it starts the next frame.
.readAgain2:
	in.h 0x300[$0], $10
	in.h 0x300[$0], $11
	and $13, $10
	and $13, $11
	cmp $10, $11
	bne .readAgain2
	cmp $10, $12
	be .readAgain2		; This line is different!

	jmp $LP

	; ----------------------------------------

pcfxKingWriteH:
	; Writes a halfword to a KING register.

	; IN:
	; $6 = register
	; $7 = value

	out.h $6, 0x600[$0]
	out.h $7, 0x604[$0]

	jmp $LP

	; ----------------------------------------

pcfxKingWriteW:
	; Writes a word to a KING register.

	; IN:
	; $6 = register
	; $7 = value

	out.h $6, 0x600[$0]
	out.w $7, 0x604[$0]

	jmp $LP

	; ----------------------------------------

pcfx7upAWrite:
	; Writes a halfword to a 7up A register.

	; IN:
	; $6 = register
	; $7 = value

	out.h $6, 0x400[$0]
	out.h $7, 0x404[$0]

	jmp $LP

	; ----------------------------------------

pcfxReadPad0:
	; Reads the status of pad 0.

	; TODO: does this only return the delta or the whole status?

	; OUT:
	; $10 = status

	; Data available?
	in.h 0[$0], $10
	andi 9, $10, $10
	cmp 1, $10	; 9 = Received data + unknown
	bne .available
	mov 0, $10
	jmp $LP

.available:
	; Read it
	mov 5, $10	; 5 = Transmit enable + receive enable
	out.h $10, 0[$0]
	in.w 0x40[$0], $10

	jmp $LP

	; ----------------------------------------

PrintStr:
	; Prints a C string to a 4-color BG in CG mode, which must start
	; at KRAM address 0 and be of the size 256*256.

	; IN:
	; $6 = X (column)
	; $7 = Y (row)
	; $8 -> string

	add -16, $SP
	st.w $LP, 0[$SP]
	st.w $20, 4[$SP]
	st.w $22, 12[$SP]

	mov $6, $20
	shl 3, $7
	st.w $7, 8[$SP]
	mov $8, $22

.charLoop:
	ld.b 0[$22], $8
	cmp 0, $8
	bz .finish
	mov $20, $6
	ld.w 8[$SP], $7
	jal PrintChar

	add 1, $20
	add 1, $22
	br .charLoop

.finish:
	ld.w 0[$SP], $LP
	ld.w 4[$SP], $20
	ld.w 12[$SP], $22
	addi 16, $SP, $SP
	jmp $LP

	; ----------------------------------------

PrintChar:
	; Prints a character from the BIOS 8*8 font to a 4-color BG in CG
	; mode, which must start at KRAM address 0 and be of the size
	; 256*256.

	; IN:
	; $6 = X (column)
	; $7 = Y (row)
	; $8 = character

	add -8, $SP
	st.w $LP, 0[$SP]
	st.w $20, 4[$SP]

	; Calculate $20 = KRAM address
	shl 6, $7
	add $7, $6
	mov $6, $20

	; Ask the BIOS for the character data address
	mov $8, $6
	mov 4, $7	; 8*8 font
	?mov 0xFFF0000C, $1
	jal .dummy
.dummy:
	add 4, $LP
	jmp $1
	; $10 now points to the character data

	mov 8, $12	; Line counter
.rowLoop:
	mov 0xD, $6	; KRAM write address
	mov $20, $7
	jal pcfxKingWriteW
	mov 0xE, $6	; KRAM R/W

	; Convert a line from 1-bit to 2-bit format
	ld.b 0[$10], $11; Source row data
	mov 0, $7	; Resulting row data
	?mov 0x80, $13	; Mask
.pixelLoop:
	shl 2, $7
	mov $11, $14	; Temporarily save...
	and $13, $11	; ... get the bit...
	mov $14, $11	; ... and restore.
	bz .dontSet
	ori 3, $7, $7	; Set the pixel
.dontSet:
	shr 1, $13
	bnz .pixelLoop

	; Write the row
	jal pcfxKingWriteH
	addi 64, $20, $20	; Move to the next BG row
	add 1, $10		; Advance to next source data row
	add -1, $12
	bnz .rowLoop

	ld.w 0[$SP], $LP
	ld.w 4[$SP], $20
	add 8, $SP
	jmp $LP